;;;; stack.scm
;;;; Kon Lovett, Mar 29 2005

;;;; Stack data structure (FIFO queue) where the value is mutable,
;;;; rather than usual pattern of the variable.

(eval-when (eval)
	(use syntax-case stack_ext_1))
	
(register-feature! 'stack)

;; Helpers

(define-syntax stack-push-1!
	(syntax-rules ()
		((_ s x) (set-cdr! s (cons x (cdr s))))))

;;;

;; MAKE-STACK
;;
;; Returns an empty stack.

(define-syntax make-stack
	(syntax-rules ()
		((_) (cons '() '()))))

;; STACK? Stack
;;
;; Stack - a stack
;;
;; Returns #t for a stack, #f otherwise. Not fool-proof!

(define-syntax stack?
	(syntax-rules ()
		((_ s) (and (list? s) (null? (car s))))))

;; STACK-EMPTY? Stack
;;
;; Stack - a stack
;;
;; Returns #t for an empty stack, #f otherwise.

(define-syntax stack-empty?
	(syntax-rules ()
		((_ s) (null? (cdr s)))))

;; STACK-LENGTH Stack
;;
;; Stack - a stack
;;
;; Returns the number of elements on the stack.

(define-syntax stack-length
	(syntax-rules ()
		((_ s) (length (cdr s)))))

;; STACK-DEPTH Stack
;;
;; Stack - a stack
;;
;; Returns the index of the last element on the stack, or -1 when stack
;; empty.

(define-syntax stack-depth
	(syntax-rules ()
		((_ s) (sub1 (stack-length s)))))

;; STACK-PEEK Stack #!optional Index
;;
;; Stack - the stack to peek into
;;
;; Index - where to peek, 0 <= & <= (stack-depth
;; Stack) - 1, defaults to 0
;;
;; Returns the peeked value.

(define-syntax stack-peek
	(syntax-rules ()
		((_ s) (car (stack-unit:node-ref 'stack-peek s 0)))
		((_ s n) (car (stack-unit:node-ref 'stack-peek s n)))))

;; STACK-POKE! Stack Value #!optional Index
;;
;; Stack - the stack to poke into
;;
;; Value - the value to poke
;;
;; Index - where to poke, 0 <= & <= (stack-depth
;; Stack) - 1, defaults to 0
;;
;; Returns the modified stack. The stack is modified in place.

(define-syntax stack-poke!
	(syntax-rules ()
		((_ s x) (begin (set-car! (stack-unit:node-ref 'stack-poke! s 0) x) s))
		((_ s x n) (begin (set-car! (stack-unit:node-ref 'stack-poke! s n) x) s))))

;; STACK-PUSH! Stack Value ...
;;
;; Stack - the stack to push onto
;;
;; Value - the new top of stack value(s)
;;
;; Returns the modified stack. The stack is modified in place.

(define-syntax stack-push!
	(syntax-rules ()
		((_ s v) (begin (stack-push-1! s v) s))
		((_ s a1 ...) (begin (for-each (lambda (x) (stack-push-1! s x)) (list a1 ...)) s))))

;; STACK-POP! Stack
;;
;; Stack - the stack to pop
;;
;; Returns the top of the stack. The stack is modified in place.

(define-syntax stack-pop!
	(syntax-rules ()
		((_ s)
			(if (stack-empty? s)
				(error 'stack-pop! "stack empty")
				(let ([x (cadr s)]) (set-cdr! s (cddr s)) x)))))

;; LIST->STACK List
;;
;; List - a proper list
;;
;; Returns the list as a stack. The resulting stack may share memory
;; with the list and the list should not be modified after this operation.

(define-syntax list->stack
	(syntax-rules ()
		((_ l) (cons '() l))))

;; STACK->LIST Stack
;;
;; Stack - a stack
;;
;; Returns the stack as a list, where the first element of the
;; list is the top element of the stack.  The resulting list may share
;; memory with the queue object and should not be modified.

(define-syntax stack->list
	(syntax-rules ()
		((_ s) (cdr s))))

;; STACK-FOR-EACH Stack Procedure
;;
;; Stack - a stack
;;
;; Procedure - a procedure of one argument
;;
;; Applies the procedure to each element of the stack, top to bottom.

(define-syntax stack-for-each
	(syntax-rules ()
		((_ f s) (for-each f (cdr s)))))

;; PRINT-STACK Stack
;;
;; Stack - a stack
;;
;; Prints the elements of the stack.

(define-syntax print-stack
	(syntax-rules ()
		((_ s) (stack-for-each print s))))
